import emcee
import corner
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from .mock import *
from .likelihoods import *

def posRange(pRanges):
    """
    Generate a position in the parameter space randomly from the prior.

    Parameters
    ----------
    pRange : list
        The ranges of the parameters.

    Returns
    -------
    m : float
        The slope.
    b : float
        The intercept.
    lnf : float (optional)
        The fraction of the model values describing the uncertainty of the model.

    Notes
    -----
    None.
    """
    ndim = len(pRanges)
    if ndim == 2:
        mR, bR = pRanges
        m = (mR[1]-mR[0]) * np.random.rand() + mR[0]
        b = (bR[1]-bR[0]) * np.random.rand() + bR[0]
        return [m, b]
    if ndim == 3:
        mR, bR, lnfR = pRanges
        m = (mR[1]-mR[0]) * np.random.rand() + mR[0]
        b = (bR[1]-bR[0]) * np.random.rand() + bR[0]
        lnf = (lnfR[1]-lnfR[0]) * np.random.rand() + lnfR[0]
        return [m, b, lnf]
    else:
        raise ValueError("[linfit]: The ndim value ({0}) is incorrect!".format(ndim))

def posBall(theta, sigma=0.01):
    """
    Generate a position in the parameter space randomly from a ball centering
    at the given paramters.

    Parameters
    ----------
    theta : float array
        The given parameters.
    sigma : float
        The radius of the ball which is the fraction of the given parameters.

    Returns
    -------
    m : float
        The slope.
    b : float
        The intercept.
    lnf : float (optional)
        The fraction of the model values describing the uncertainty of the model.

    Notes
    -----
    None.
    """
    ndim = len(theta)
    p = np.array(theta) * (1 + (np.random.rand(ndim)-0.5)*2*sigma)
    return p

class LinFit(object):
    """
    Linear regression with MCMC method. The MCMC backend is emcee.
    The y of the data could be upperlimits. The upper limits are properly deal
    with using the method mentioned by Sawicki (2012).
    """
    def __init__(self, x, y, pRanges, xerr=None, yerr=None, flag=None, nsigma=3,
                 lnlikeType="Nukers"):
        """
        To initiate the object.

        Parameters
        ----------
        x : float array
            The x data.
        y : float array
            The y data.
        pRanges : list
            The ranges of the linear model.
            [
              [m(min), m(max)],
              [b(min), b(max)],
              [lnf(min), lnf(max)](optional)
            ]
        xerr : float array
            The uncertainty of x data.
        yerr : float array
            The uncertainty of y data.
        flag : float array
            The upperlimit flag of the y data; 0 for the detection and 1 for the
            upperlimit.
        nsigma : float
            The upperlimits are nsigma times of the uncertainties, respectively.

        Notes
        -----
        None.
        """
        self.x = x
        self.y = y
        self.pRanges = pRanges
        self.flag = flag
        self.nsigma = nsigma
        ndim = len(pRanges)
        if ndim == 2:
            print("[linfit]: The model uncertainty is NOT considered!")
        elif ndim == 3:
            print("[linfit]: The model uncertainty is considered!")
        else:
            raise ValueError("[linfit]: The parameter number ({0}) is incorrect!".format(ndim))
        self.ndim = ndim
        if (xerr is None) & (yerr is None):
            xerr = np.zeros_like(x)
            yerr = np.ones_like(y)
        else:
            if xerr is None:
                xerr = np.zeros_like(x)
            if yerr is None:
                yerr = np.zeros_like(y)
        self.xerr = xerr
        self.yerr = yerr
        if lnlikeType == "Nukers":
            self.lnlike = lnlike_Nukers
            self.parNames = ["beta", "alpha", "epsy0"]
        elif lnlikeType == "gcs":
            self.lnlike = lnlike_gcs
            self.parNames = ["m", "b", "lnf"]
        elif lnlikeType == "naive":
            self.lnlike = lnlike_naive
            self.parNames = ["m", "b", "lnf"]
        elif lnlikeType == "perp":
            self.lnlike = lnlike_perp
            self.parNames = ["theta", "bv", "V"]
        elif lnlikeType == "perp2":
            self.lnlike = lnlike_perp2
            self.parNames = ["theta", "b", "V"]
        else:
            raise ValueError("[linfit]: The lnlike function ({0}) is not recognised!".format(lnlike))
        self.lnlikeType = lnlikeType

    def lnprob(self, theta, x, y, xerr, yerr, pRanges, *args, **kwargs):
        """
        The ln of probability function.

        Parameters
        ----------
        lnlike : function
            The lnlike function
        theta : list
            The list of the model parameters, [m, b, lnf (optional)].
        x : float array
            The data of x.
        y : float array
            The data of y.
        xerr : float array
            The uncertainty of the x data.
        yerr : float array
            The uncertainty of the y data.
        pRanges : list
            The list of the parameter prior ranges.
        args and kwargs : for the ChiSq function.

        Returns
        -------
        The ln probability.

        Notes
        -----
        None.
        """
        lp = lnprior(theta, pRanges)
        if not np.isfinite(lp):
            return -np.inf
        return lp + self.lnlike(theta, x, y, xerr, yerr, *args, **kwargs)

    def EnsembleSampler(self, nwalkers, **kwargs):
        """
        Set up the EnsembleSampler of the emcee.
        """
        self.nwalkers = nwalkers
        self.__sampler = "EnsembleSampler"
        if self.lnlikeType == "gcs":
            logpargs = (self.x, self.y, self.xerr, self.yerr, self.pRanges,
                    self.flag, self.nsigma)
        else:
            logpargs = (self.x, self.y, self.xerr, self.yerr, self.pRanges)
        self.sampler = emcee.EnsembleSampler(nwalkers, self.ndim, self.lnprob,
                       args=logpargs, **kwargs)
        print("[linfit]: Use the EnsembleSampler.")
        return self.sampler

    def PTSampler(self, ntemps, nwalkers, **kwargs):
        self.ntemps = ntemps
        self.nwalkers = nwalkers
        self.__sampler = "PTSampler"
        if self.lnlikeType == "gcs":
            loglargs = (self.x, self.y, self.xerr, self.yerr, self.flag, self.nsigma)
        else:
            loglargs = (self.x, self.y, self.xerr, self.yerr)
        self.sampler = emcee.PTSampler(ntemps, nwalkers, self.ndim,
                       logl=self.lnlike, logp=lnprior,
                       loglargs=loglargs, logpargs=[self.pRanges], **kwargs)
        print("[linfit]: Use the PTSampler.")
        return self.sampler

    def run_mcmc(self, p, nrun):
        """
        Run the MCMC sampling.

        Parameters
        ----------
        p : float array
            The initial positions of the walkers.
        nrun : float
            The steps to run the sampling.

        Returns
        -------
        None.

        Notes
        -----
        None.
        """
        samplerType = self.__sampler
        nwalkers = self.nwalkers
        ndim = self.ndim
        p = np.array(p)
        if samplerType == "EnsembleSampler":
            assert p.shape == (nwalkers, ndim)
        elif samplerType == "PTSampler":
            assert p.shape == (self.ntemps, nwalkers, ndim)
        sampler = self.sampler
        sampler.run_mcmc(p, nrun)

    def reset(self):
        self.sampler.reset()

    def p_prior(self):
        """
        Generate the positions of the walkers in the parameter space randomly
        from the prior.

        Parameters
        ----------
        None.

        Returns
        -------
        p : float array
            The positions of the walkers.

        Notes
        -----
        None.
        """
        sampler  = self.__sampler
        nwalkers = self.nwalkers
        pRanges = self.pRanges
        if sampler == "EnsembleSampler":
            p = [posRange(pRanges) for i in range(nwalkers)]
        elif sampler == "PTSampler":
            ntemps = self.ntemps
            p = np.zeros((ntemps, nwalkers, self.ndim))
            for loop_t in range(ntemps):
                for loop_w in range(nwalkers):
                    p[loop_t, loop_w, :] = posRange(pRanges)
        return p

    def p_ball(self, theta, sigma=0.01):
        """
        Generate the position of the walkers in the parameter space randomly from
        a ball centering at the given paramters.

        Parameters
        ----------
        theta : float array
            The given parameters.
        sigma : float
            The radius of the ball which is the fraction of the given parameters.

        Returns
        -------
        p : float array
            The positions of the walkers.

        Notes
        -----
        None.
        """
        sampler  = self.__sampler
        nwalkers = self.nwalkers
        if sampler == "EnsembleSampler":
            p = [posBall(theta) for i in range(nwalkers)]
        elif sampler == "PTSampler":
            ntemps = self.ntemps
            p = np.zeros((ntemps, nwalkers, self.ndim))
            for loop_t in range(ntemps):
                for loop_w in range(nwalkers):
                    p[loop_t, loop_w, :] = posBall(theta)
        return p

    def p_logl_max(self):
        """
        Find the position in the sampled parameter space that the likelihood is
        the highest.
        """
        sampler = self.__sampler
        if sampler == "EnsembleSampler":
            chain  = self.sampler.chain
            lnlike = self.sampler.lnprobability
        elif sampler == "PTSampler":
            chain  = self.sampler.chain[0, ...]
            lnlike = self.sampler.lnlikelihood[0, ...]
        else:
            raise ValueError("[linfit]: The sampler type ({0}) is unrecognised!".format(sampler))
        idx = lnlike.ravel().argmax()
        p   = chain.reshape(-1, self.ndim)[idx]
        return p

    def fit(self, nrun=1000, nburnin=500, psigma=0.01):
        self.__nrun = nrun
        self.__nburnin = nburnin
        ndim = self.ndim
        p = self.p_prior()
        self.run_mcmc(p, nburnin)
        pmax = self.p_logl_max()
        p = self.p_ball(pmax, psigma)
        self.reset()
        self.run_mcmc(p, nrun)

    def posterior_sample(self, burnin=0):
        """
        Return the samples merging from the chains of all the pertinent walkers.

        Parameters
        ----------
        burnin : float
            The number of initial samplings that would be dropped as burn-in run.

        Returns
        -------
        samples : float array
            The samples of the parameter space according to the posterior probability.

        Notes
        -----
        None.
        """
        self.__burnin = burnin
        sampler  = self.sampler
        nwalkers = self.nwalkers
        if self.__sampler == "EnsembleSampler":
            chain = sampler.chain
            lnprob = sampler.lnprobability[:, -1]
        elif self.__sampler == "PTSampler":
            chain = np.squeeze(sampler.chain[0, ...])
            lnprob = np.squeeze(sampler.lnprobability[0, :, -1])
        samples = chain[:, burnin:, :].reshape((-1, self.ndim))
        return samples

    def get_BestFit(self, burnin=0):
        parNames = self.parNames
        ndim = self.ndim
        samples = self.posterior_sample(burnin)
        lnlikeType = self.lnlikeType
        BFDict = {
            "Parameters": parNames,
            "samples": samples
        }
        bfList = map(lambda v: (v[1], v[2]-v[1], v[1]-v[0]),
                     zip(*np.percentile(samples, [16, 50, 84], axis=0)))
        fmerge = lambda v: (v[0]-v[2], v[0], v[0]+v[1])
        fdiff  = lambda v: (v[1], v[2]-v[1], v[1]-v[0])
        for loop in range(ndim):
            pn = parNames[loop]
            BFDict[pn] = bfList[loop]
        if lnlikeType == "naive" or lnlikeType == "gcs":
            BFDict["slope"] = BFDict["m"]
            BFDict["intercept"] = BFDict["b"]
            lnf = fmerge(BFDict["lnf"])
            BFDict["IntSc"] = fdiff(np.exp(lnf))
        elif lnlikeType == "Nukers":
            BFDict["slope"] = BFDict["beta"]
            BFDict["intercept"] = BFDict["alpha"]
            BFDict["IntSc"] = BFDict["epsy0"]
        elif lnlikeType == "perp":
            t = fmerge(BFDict["theta"])
            BFDict["slope"] = fdiff(np.tan(t))
            bv = fmerge(BFDict["bv"])
            BFDict["intercept"] = fdiff(bv / np.cos(t))
            V = fmerge(BFDict["V"])
            BFDict["IntSc"] = fdiff(np.sqrt(V) / np.cos(t))
        elif lnlikeType == "perp2":
            t = fmerge(BFDict["theta"])
            BFDict["slope"] = fdiff(np.tan(t))
            BFDict["intercept"] = BFDict["b"]
            V = fmerge(BFDict["V"])
            BFDict["IntSc"] = fdiff(np.sqrt(V) / np.cos(t))
        return BFDict

    def get_lnprob(self, theta):
        if self.lnlikeType == "gcs":
            logpargs = (self.x, self.y, self.xerr, self.yerr, self.pRanges,
                    self.flag, self.nsigma)
        else:
            logpargs = (self.x, self.y, self.xerr, self.yerr, self.pRanges)
        lnP = self.lnprob(theta, *logpargs)
        return lnP

    def get_BFProb(self, burnin=0):
        sampler = self.__sampler
        BFDict = self.get_BestFit(burnin)
        parNames = self.parNames
        theta = []
        for pn in parNames:
            theta.append(BFDict[pn][0])
        lnP = self.get_lnprob(theta)
        return lnP

    def plot_BestFit(self, burnin=0):
        #Plot the data
        fig = plt.figure()
        x = self.x
        y = self.y
        xerr = self.xerr
        yerr = self.yerr
        flag = self.flag
        if flag is None:
            plt.errorbar(x, y, xerr=xerr, yerr=yerr, linestyle="none", marker=".",
                         color="k")
        else:
            fltr = flag == 0
            plt.errorbar(x[fltr], y[fltr], xerr=xerr[fltr], yerr=yerr[fltr],
                         fmt=".k")
            fltr = flag == 1
            plt.errorbar(x[fltr], y[fltr], xerr=xerr[fltr], yerr=0.5, uplims=True,
                         fmt=".k")
        ax = plt.gca()
        ax.tick_params(axis='both', labelsize=20)
        xl = np.array(ax.get_xlim())
        #Plot the best fit
        BFDict = self.get_BestFit(burnin)
        m_fit = BFDict["slope"]
        b_fit = BFDict["intercept"]
        m_bf = m_fit[0]
        b_bf = b_fit[0]
        yl = m_bf * xl + b_bf
        plt.plot(xl, yl, color="k", linestyle="--", label="{0}".format("Best Fit"))
        return (fig, ax)

    def plot_corner(self, burnin=0):
        BFDict = self.get_BestFit(burnin)
        samples = BFDict["samples"]
        fig = corner.corner(samples, labels=self.parNames,
                            quantiles=[0.16, 0.5, 0.84], show_titles=True,
                            title_kwargs={"fontsize": 12})
        ax = plt.gca()
        return (fig, ax)

    def plot_chain(self):
        ndim = self.ndim
        samplerType = self.__sampler
        sampler = self.sampler
        parNames = self.parNames
        if samplerType == "EnsembleSampler":
            chain  = self.sampler.chain
        elif samplerType == "PTSampler":
            chain  = self.sampler.chain[0, ...]
        fig, axes = plt.subplots(ndim, 1, sharex=True, figsize=(8, 9))
        axes[0].plot(chain[:, :, 0].T, color="k", alpha=0.4)
        axes[0].yaxis.set_major_locator(MaxNLocator(5))
        axes[0].set_ylabel(parNames[0], fontsize=24)
        axes[1].plot(chain[:, :, 1].T, color="k", alpha=0.4)
        axes[1].yaxis.set_major_locator(MaxNLocator(5))
        axes[1].set_ylabel(parNames[1], fontsize=24)
        if ndim == 3:
            axes[2].plot(chain[:, :, 2].T, color="k", alpha=0.4)
            axes[2].yaxis.set_major_locator(MaxNLocator(5))
            axes[2].set_ylabel(parNames[2], fontsize=24)
            axes[2].set_xlabel("step number", fontsize=24)
        fig.tight_layout(h_pad=0.0)
        return (fig, axes)



import numpy as np
import matplotlib.pyplot as plt

def MockData(m, b, lnf=None, sigy=None, x=None, ndata=None, xep=[0.0, 0.0], yep=[0.0, 0.0]):
    """
    Generate the mock data.

    Parameters
    ----------
    m : float
        The slope.
    b : float
        The intercept.
    lnf : float, optional (not with sigy)
        The intrinsic scatter as a fraction of the model. The default is None.
    sigy : float, optional (not with lnf)
        The intrinsic scatter as the dispersion in Y axis. The default is None.
    x : float array, optional
        The x of the data. The default is None.
    ndata : float, optional
        If x is not provided, ndata should be given to provide the number of the
        data randomly generated. The default is None.
    xep : list
        The lower and upper boundaries of the x error. The uncertainties are assumed
        to be Gaussian. The default is [0.0, 0.0].
    yep : list
        The lower and upper boundaries of the y error. The uncertainties are assumed
        to be Gaussian. The default is [0.0, 0.0].
    """
    if x is None:
        assert not ndata is None
        x = np.sort(10*np.random.rand(ndata))
    else:
        if (not ndata is None) & (ndata != len(x)):
            raise Warning("[linfit]: The input ndata ({0}) is not used.".format(ndata))
        ndata = len(x)
    xerr = xep[0] + (xep[1] - xep[0]) * np.random.rand(ndata)
    yerr = yep[0] + (yep[1] - yep[0]) * np.random.rand(ndata)
    x += xerr * np.random.randn(ndata)
    plt.figure()
    y = m * x + b
    plt.plot(x, y, linestyle="none", marker=".", label="origin")
    if not lnf is None:
        assert sigy is None
        f = np.exp(lnf)
        y += np.abs(f * y) * np.random.randn(ndata)
        plt.plot(x, y, linestyle="none", marker=".", label="intr. scatt.")
    if not sigy is None:
        assert lnf is None
        y += sigy * np.random.randn(ndata)
        plt.plot(x, y, linestyle="none", marker=".", label="intr. scatt.")
    y += yerr * np.random.randn(ndata)
    plt.plot(x, y, linestyle="none", marker=".", label="measure")
    plt.legend(fontsize=18)
    retDict = {
        "x": x,
        "y": y,
        "xerr": xerr,
        "yerr": yerr
    }
    return retDict

def MockData_perp(theta, b, sqrtV=None, x0=None, ndata=None, xep=[0.0, 0.0], yep=[0.0, 0.0]):
    """
    Generate the mock data with intrinsic scatter perpendicular to the line.

    Parameters
    ----------
    theta : float
        The slope.
    b : float
        The intercept.
    sqrtV : float, optional
        The intrinsic scatter perpendicular to the line. The default is None.
    x0 : float array, optional
        The x of the data along the line. The default is None.
    ndata : float, optional
        If x is not provided, ndata should be given to provide the number of the
        data randomly generated. The default is None.
    xep : list
        The lower and upper boundaries of the x error. The uncertainties are assumed
        to be Gaussian. The default is [0.0, 0.0].
    yep : list
        The lower and upper boundaries of the y error. The uncertainties are assumed
        to be Gaussian. The default is [0.0, 0.0].
    """
    if x0 is None:
        assert not ndata is None
        x0 = np.sort(10*np.random.rand(ndata))
    else:
        if (not ndata is None) & (ndata != len(x)):
            raise Warning("[linfit]: The input ndata ({0}) is not used.".format(ndata))
        ndata = len(x0)
    #Make dispersion perpendicular to the line
    y0 = sqrtV * np.random.randn(ndata)
    #Rotate the line
    vecs = np.array([x0, y0])
    rotMatrix = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
    vecsR = np.dot(rotMatrix, vecs)
    x = vecsR[0, :]
    y = vecsR[1, :] + b
    #Add measurement uncertainty
    xerr = xep[0] + (xep[1] - xep[0]) * np.random.rand(ndata)
    yerr = yep[0] + (yep[1] - yep[0]) * np.random.rand(ndata)
    x += xerr * np.random.randn(ndata)
    y += yerr * np.random.randn(ndata)
    retDict = {
        "x": x,
        "y": y,
        "xerr": xerr,
        "yerr": yerr
    }
    return retDict



import numpy as np
import matplotlib.pyplot as plt
from linfit import LinFit

# Reproducible results!
np.random.seed(123)

# Choose the "true" parameters.
m_true = -0.9594
b_true = 4.294
f_true = 0.534

# Generate some synthetic data from the model.
N = 50
x = np.sort(10*np.random.rand(N))
yerr = 0.1+0.5*np.random.rand(N)
y = m_true*x+b_true
y += np.abs(f_true*y) * np.random.randn(N)
y += yerr * np.random.randn(N)

# Plot the dataset and the true model.
xl = np.array([0, 10])
plt.errorbar(x, y, yerr=yerr, fmt=".k")
plt.plot(xl, m_true*xl+b_true, "k", lw=3, alpha=0.6)
plt.ylim(-9, 9)
plt.xlabel("$x$")
plt.ylabel("$y$")
plt.tight_layout()
plt.savefig("line-data.png")
plt.close()

pRanges = [[-10, 10], [-20, 20], [-10, 1]]
lf = LinFit(x, y, pRanges, yerr=yerr)
lf.EnsembleSampler(nwalkers=100)
#lf.PTSampler(ntemps=10, nwalkers=100)
print("Start fitting!")
lf.fit(nrun=1000, nburnin=500, psigma=0.01)
print("Finish fitting!")
burnin = 200
BFDict = lf.get_BestFit(burnin)

lf.plot_BestFit(burnin)
plt.savefig("best-fit.png")
plt.close()

lf.plot_corner(burnin)
plt.savefig("corner.png")
plt.close()

lf.plot_chain()
plt.savefig("chain.png")
plt.close()



import numpy as np
import matplotlib.pyplot as plt
from linfit import MockData

# Reproducible results!
np.random.seed(123)

# Choose the "true" parameters.
m_true = -0.9594
b_true = 4.294
sigy_true = 1.303

# Generate some synthetic data from the model.
N = 50
x = np.sort(10*np.random.rand(N))
retDict = MockData(m_true, b_true, sigy=sigy_true, x=x, ndata=N, xep=[0.1, 0.6], yep=[0.1, 0.6])
y = retDict["y"]
xerr = retDict["xerr"]
yerr = retDict["yerr"]
flag = np.zeros(N, dtype=bool)
#lou = [3, 9, 17]
lou = []
for loop in lou:
    flag[loop] = 1
    print y[loop]
    while(3*yerr[loop] < y[loop]):
        yerr[loop] *= 1.5
    y[loop] = yerr[loop] * 3
    print y[loop]

#Save the data
scp = "sigy" #"upp"
# Plot the dataset and the true model.
xl = np.array([0, 10.8])
uflag = flag == 0
plt.errorbar(x[uflag], y[uflag], xerr=xerr[uflag], yerr=yerr[uflag], fmt=".k")
plt.errorbar(x[flag], y[flag], xerr=xerr[flag], yerr=0.5, uplims=True, fmt=".k")
plt.plot(xl, m_true*xl+b_true, "k", lw=3, alpha=0.6)
plt.xlim(0, 10.8)
plt.ylim(-10, 10)
plt.xlabel("$x$")
plt.ylabel("$y$")
plt.tight_layout()
plt.savefig("line-data_{0}.png".format(scp))
plt.close()

data = np.array([x, y, xerr, yerr, flag])
fname = "data_{0}.txt".format(scp)
np.savetxt(fname, data.T, fmt="%.4f")



from distutils.core import setup

setup(
  name = 'linfit',
  packages = ['linfit'], # this must be the same as the name above
  version = '1.0.1',
  description = 'Linear regression with MCMC',
  author = 'Jinyi Shangguan',
  author_email = 'shangguan@pku.edu.cn',
  license="MIT",
  url = 'https://github.com/darkbear9494/LinFit',
  download_url = 'https://github.com/darkbear9494/LinFit/tarball/1.0',
  keywords = ['fitting', 'MCMC', 'linear regression'],
  install_requires=["numpy", "matplotlib", "emcee"],
  classifiers = [],
)



import numpy as np
import matplotlib.pyplot as plt

theta = 0.25 * np.pi
"""
ndata = 50
sqrtV = 0.4
x0 = np.sort(10*np.random.rand(ndata))
y0 = sqrtV * np.random.randn(ndata)
vecs = np.array([x0, y0])
rotMatrix = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
vecsR = np.dot(rotMatrix, vecs)
x1 = vecsR[0, :]
y1 = vecsR[1, :]
plt.plot(x0, y0, linestyle="none", marker=".")
plt.plot(x1, y1, linestyle="none", marker=".")
plt.show()
"""

print np.log(np.e)
print np.cos(np.pi/3)



import numpy as np
from scipy.special import erf

#--> The lnlikes
#-> The lnlike calculated with generalized chi square
#The generalized chi-square function
def ChiSq(data, model, unct=None, flag=None, nsigma=3):
    '''
    This is a generalized chi-square function that allows y to be upperlimits. The
    upper limits are properly deal with using the method mentioned by Sawicki (2012).

    Parameters
    ----------
    data : float array
        The observed data and upperlimits.
    model : float array
        The model.
    unct : float array or Nobe by default
        The uncertainties.
    flag : float array or None by default
        The flag of upperlimits, 0 for detection and 1 for upperlimits.
    nsigma : float; default: 3
        The provided upperlimits are nsigma times of the uncertainties.

    Returns
    -------
    chsq : float
        The Chi square

    Notes
    -----
    This chi-square form consider the x and y asymmetrically except for some special
    situations.
    '''
    if unct is None:
        unct = np.ones_like(data)
    if flag is None:
        flag = np.zeros_like(data)
    fltr_dtc = flag == 0
    fltr_non = flag == 1
    if np.sum(fltr_dtc)>0:
        wrsd_dtc = (data[fltr_dtc] - model[fltr_dtc])/unct[fltr_dtc] #The weighted residual
        chsq_dtc = np.sum(wrsd_dtc**2) + np.sum( np.log(2 * np.pi * unct[fltr_dtc]**2.0) )
    else:
        chsq_dtc = 0.
    if np.sum(fltr_non)>0:
        unct_non = data[fltr_non]/nsigma #The nondetections are 3 sigma upper limits.
        wrsd_non = (data[fltr_non] - model[fltr_non])/(unct_non * 2**0.5)
        chsq_non = np.sum( -2.* np.log( 0.5 * (1 + erf(wrsd_non)) ) )
    else:
        chsq_non = 0.
    chsq = chsq_dtc + chsq_non
    return chsq

def lnlike_gcs(theta, x, y, xerr, yerr, *args, **kwargs):
    """
    The ln of likelihood function use the generalized chi-square function. The y
    of the data could be upperlimits.

    Parameters
    ----------
    theta : list
        The list of the model parameters, [m, b, lnf (optional)].
    x : float array
        The data of x.
    y : float array
        The data of y.
    xerr : float array
        The uncertainty of the x data.
    yerr : float array
        The uncertainty of the y data.
    args and kwargs : for the ChiSq function.

    Returns
    -------
    The ln likelihood.

    Notes
    -----
    None.
    """
    if len(theta) == 2:
        m, b = theta
        model = m * x + b
        s = np.sqrt(yerr**2 + (m*xerr)**2)
    if len(theta) == 3:
        m, b, lnf = theta
        model = m * x + b
        s = np.sqrt(yerr**2 + (m*xerr)**2 + model**2*np.exp(2*lnf))
    lnL = -0.5 * ChiSq(y, model, s, *args, **kwargs)
    return lnL

#-> The Nukers' lnlike
def lnlike_Nukers(theta, x, y, epsx, epsy):
    """
    This is the ln likelihood function resembling the Nukers' estimate from
    Tremaine et al. (2002). One of the merits of this form is that the x and y
    are symmetric (see the paper for the details). The symbols of the parameters
    also follow the paper.

    Parameters
    ----------
    theta : list
        The list of model parameters, [beta, alpha, epsy0 (optional)].
        beta is the slope, alpha is the intercept and epsy0 is the intrinsic
        scatter along the y direction.
    x : float array
        The x data.
    y : float array
        The y data.
    epsx : float array
        The uncertainty of x data.
    epsy : float array
        The uncertainty of y data.

    Returns
    -------
    lnL : float
        The ln of the likelihood.

    Notes
    -----
    The lnlike penalizes the very broad intrinsic dispersion assuming it is a
    Gaussian distribution. Therefore, the optimization is to seek the maximum of
    the lnlike instead of the Nukers' estimate ~1.
    """
    if len(theta) == 2:
        beta, alpha = theta
        inv_sigma2 = 1.0/(epsy**2 + (beta * epsx)**2)
    if len(theta) == 3:
        beta, alpha, epsy0 = theta
        inv_sigma2 = 1.0/(epsy**2 + epsy0**2 + (beta * epsx)**2)
    lnL = -0.5*(np.sum((y - alpha - beta * x)**2*inv_sigma2 - np.log(inv_sigma2)))
    return lnL

#-> The lnlike calculated from the distance perpendicular to the line following
#Hogg et al. (2010; arXiv:1008.4686)
def lnlike_perp(theta, x, y, sigx, sigy):
    """
    This is the ln likelihood function considering the 2-dimensional uncertainties
    and calculated based on the distance of the points perpendicular to the line.
    It follows the equation (35) of Hogg et al. (2010; arXiv:1008.4686).

    Parameters
    ----------
    theta : list
        The list of model parameters, [t, bv, V (optional)]. t (in radian) is
        the angle (theta = arctan slope), bv is the perpendicular distance of
        the line from the origin and V is intrinsic Gaussian variance orthogonal
        to the line.
    x : float array
        The x data.
    y : float array
        The y data.
    sigx : float array
        The uncertainty of x data.
    sigy : float array
        The uncertainty of y data.

    Returns
    -------
    lnL : float
        The ln likelihood.

    Notes
    -----
    None.
    """
    if len(theta) == 2:
        t, bv = theta
        V = 0
    if len(theta) == 3:
        t, bv, V = theta
    delta = y * np.cos(t) - x * np.sin(t) - bv
    Sigma2 = (sigx * np.sin(t))**2 + (sigy * np.cos(t))**2
    lnL = -0.5 * np.sum(delta**2 / (Sigma2 + V) + np.log(Sigma2 + V))
    return lnL

def lnlike_perp2(theta, x, y, sigx, sigy):
    """
    This is the ln likelihood function considering the 2-dimensional uncertainties
    and calculated based on the distance of the points perpendicular to the line.
    It follows the equation (35) of Hogg et al. (2010; arXiv:1008.4686).

    Parameters
    ----------
    theta : list
        The list of model parameters, [t, b, V (optional)]. t (in radian) is
        the angle (theta = arctan slope), b is the intercept and V is intrinsic
        Gaussian variance orthogonal
        to the line.
    x : float array
        The x data.
    y : float array
        The y data.
    sigx : float array
        The uncertainty of x data.
    sigy : float array
        The uncertainty of y data.

    Returns
    -------
    lnL : float
        The ln likelihood.

    Notes
    -----
    None.
    """
    if len(theta) == 2:
        t, b = theta
        V = 0
    if len(theta) == 3:
        t, b, V = theta
    delta = (y - b) * np.cos(t) - x * np.sin(t)
    Sigma2 = (sigx * np.sin(t))**2 + (sigy * np.cos(t))**2
    lnL = -0.5 * np.sum(delta**2 / (Sigma2 + V) + np.log(Sigma2 + V))
    return lnL

#-> The lnlike that considers the model imperfectness naively as a fraction of
#the model values.
def lnlike_naive(theta, x, y, xerr, yerr):
    """
    The ln of likelihood function using all detected data.

    Parameters
    ----------
    theta : list
        The list of the model parameters, [m, b, lnf (optional)].
    x : float array
        The data of x.
    y : float array
        The data of y.
    xerr : float array
        The uncertainty of the x data.
    yerr : float array
        The uncertainty of the y data.

    Returns
    -------
    The ln likelihood.

    Notes
    -----
    None.
    """
    if len(theta) == 2:
        m, b = theta
        model = m * x + b
        inv_sigma2 = 1.0/(yerr**2 + (m*xerr)**2)
        return -0.5*(np.sum((y-model)**2*inv_sigma2 - np.log(inv_sigma2)))
    if len(theta) == 3:
        m, b, lnf = theta
        model = m * x + b
        inv_sigma2 = 1.0/(yerr**2 + (m*xerr)**2 + model**2*np.exp(2*lnf))
        return -0.5*(np.sum((y-model)**2*inv_sigma2 - np.log(inv_sigma2)))
    else:
        raise ValueError("[linfit]: The length of parameters ({0}) is incorrect!".format(len(theta)))

def lnprior(theta, pRanges):
    """
    The ln of prior function.

    Parameters
    ----------
    theta : list
        The list of the model parameters, [m, b, lnf (optional)].
    pRanges : list
        The list of the parameter prior ranges.

    Returns
    -------
    The ln prior.

    Notes
    -----
    None.
    """
    assert len(theta) == len(pRanges)
    if len(theta) == 2:
        m, b = theta
        mR, bR = pRanges
        if mR[0] < m < mR[1] and bR[0] < b < bR[1]:
            return 0.0
        return -np.inf
    if len(theta) == 3:
        m, b, lnf = theta
        mR, bR, lnfR = pRanges
        if mR[0] < m < mR[1] and bR[0] < b < bR[1] and lnfR[0] < lnf < lnfR[1]:
            return 0.0
        return -np.inf
    else:
        raise ValueError("[linfit]: The length of parameters ({0}) is incorrect!".format(len(theta)))

def lnprob(theta, x, y, xerr, yerr, pRanges, *args, **kwargs):
    """
    The ln of probability function.

    Parameters
    ----------
    theta : list
        The list of the model parameters, [m, b, lnf (optional)].
    x : float array
        The data of x.
    y : float array
        The data of y.
    xerr : float array
        The uncertainty of the x data.
    yerr : float array
        The uncertainty of the y data.
    pRanges : list
        The list of the parameter prior ranges.
    args and kwargs : for the ChiSq function.

    Returns
    -------
    The ln probability.

    Notes
    -----
    None.
    """
    lp = lnprior(theta, pRanges)
    if not np.isfinite(lp):
        return -np.inf
    return lp + lnlike(theta, x, y, xerr, yerr, *args, **kwargs)
